monitoring_report_base = "/Report/Accountability/"
monitoring_links = {
    "15.1": "15_1_-_Cohort_Outcome_-_Counts_and_Rates",
    "15.2": "15_2_-_Cohort_Outcome_-_Student_Details",
    "16.7": "16_7_-_Students_with_Disabilities_-_Monitoring_Counts_ODS",
    "16.8": "16_8_-_Students_with_Disabilities_-_Monitoring_Student_List_ODS",
}

snapshot_report_base = "/Report/Snapshot/"
snapshot_links = {
    "Fall1": {
        "1.1": "1_1_EnrollmentPrimaryStatusbySubgroup",
        "1.2": "1_2_EnrollmentPrimaryStatusStudentList",
        "1.3": "1_3_EnrollmentPrimaryStatusDisaggregated",
        "1.4": "1_5_EnrollmentByStatusDisaggregated",
        "1.5": "1_5_EnrollmentByStatusDisaggregated",
        "1.8": "1_8_DropoutsBySubgroupStudentList",
        "1.12": "1_12_DropoutsDisaggregated",
        "1.13": "1_13_ExitCountsDisaggregated",
        "1.14": "1_14_DropoutCountsStateView",
        "1.17": "1_17_FRPM_EnglishLearnerFosterYouthCount_AY23",
        "1.18": "1_18_FPRM_EnglishLearnerFosterYouthStudentList_AY23",
        "1.24": "1_24_DropoutsBySubgroup",
        "2.1": "2_1_TitleIIIEligibleImmigrantCounts",
        "2.2": "2_2_TitleIIIEligibleImmigrantsCountsByBirthCountry",
        "2.8": "2_8_ELAStatusCountByPrimaryLanguage_Fall_",
        "2.9": "2_9_ELASCensusComparison_Fall1_",
        "2.10": "2_10_EnglishLanguageAcquisitionStatusCensusComparisonStudentList",
        "2.14": "2_14_LanguageInstructionProgramCount",
        "2.15": "2_15_LanguageInstruction_StudentList",
        "8.1": "8_1_StudentProfileList",
        "8.1a": "8_1a_StudentProfileExitsList",
        "8.1b": "8_1b_StudentProfileDropoutsList",
        "8.1c": "8_1c_StudentProfileDropouts_StateView",
        "16.1": "16_1_SPEDEducationPlanByPrimaryDisabilityCount",
        "16.2": "16_2_SPEDPrimaryDisabilityByFederalSettingCount",
        "16.3": "16_3_SPEDStudentProfileList",
        "16.5": "16_5_StudentsWithDisabilitiesStudentServices",
        "16.6": "16_6_StudentsWithDisabilitiesStudentServicesStudentList",
        "16.11": "16_11_SPEDPrimaryDisabilityAnnualComparisionReport",
    },
    "Fall2": {
        "2.4": "2_4_EnglishLearnerEducationServicesStudentCountsUnduplicated",
        "2.5": "2_5_EnglishLearnerEducationServicesUnduplicatedCountOfTeachersProvidingELServices",
        "2.6": "2_6_EnglishLearnerCountByPrimaryLanguage",
        "2.7": "2_7_EnglishLearnerEducationServicesStudentList",
        "2.11": "2_11_EnglishLearnerInstructionalPlanCountByPlan",
        "3.1": "3_1_ClassSizeByContentArea",
        "3.2": "3_2_ClassEnrollmentByContentArea",
        "3.3": "3_3_ClassEnrollmentStudentList",
        "3.6": "3_6_CourseSectionEnrollmentCountsByContentArea",
        "3.7": "3_7_CourseSectionEnrollmentCountAndDetails",
        "3.8": "3_8_CourseSectionEnrollmentStudentList",
        "4.1": "4_1_StaffCountFTEByJobClassification",
        "4.2": "4_2_StaffCountAndFTEbyJobClassificationDisaggregated",
        "4.3": "4_3_StaffTeachingAssignmentsDetail",
        "4.4": "4_4_StaffProfileList",
        "4.5": "4_5_StaffNonClassroomBasedorSupportAssignmentDetail",
        "17.1": "17_1_PostsecondaryStatusOutcomeCount",
        "17.2": "17_2_PostsecondaryStatusOutcomeStudentList",
    },
    "EOY1": {
        "3.9": "3_9_CourseSectionsCompletedCountByContentAreaForDepartmentalizedCourses",
        "3.10": "3_10_CourseSectionsCompletedCountAndDetailsForDepartmentalizedCourses",
        "3.11": "3_11_CourseSectionsCompletedStudentListForDepartmentalizedCourses",
        "3.14": "3_14_CareerTechnicalEducationParticipantsCount",
        "3.15": "3_15_CareerTechnicalEducationParticipantsStudentList",
        "3.16": "3_16_EducationalOptionsCourseCompletionStudentCount",
        "3.19": "3_19_CareerTechnicalEducationCompletersCountbyPathway",
        "3.20": "3_20_CareerTechnicalEducationCompletersStudentList",
        "18.1": "18_1_WorkBasedLearningCount",
        "18.2": "18_2_WorkBasedLearningStudentList",
    },
    "EOY2": {
        "5.1": "5_1_ProgramParticipantsCount",
        "5.2": "5_2_ProgramParticipantsNCLBTitleIPartABasicTargetedEducationServices",
        "5.3": "5_3_ProgramParticipantsStudentList",
    },
    "EOY3": {
        "1.21": "1_21_CumulativeEnrollmentCount",
        "1.22": "1_22_GraduatesandCompletersCount",
        "1.23": "1_23_GraduatesandCompletersStudentList",
        "2.16": "2_16_EnglishLanguageAcquisitionStatusELsReclassifiedRFEP_EOY3",
        "2.17": "2_17_EnglishLanguageAcquisitionStatusELsReclassifiedRFEPStudentList_EOY3",
        "5.4": "5_4_HomelessStudentsEnrolledUnduplicatedCountbySchool",
        "5.5": "5_5_HomelessStudentList",
        "7.10": "7_10_IncidentCount",
        "7.11": "7_11_IncidentResultCount",
        "7.12": "7_12_IncidentResultsStudentList",
        "7.13": "7_13_StudentOffense_CountByOffense",
        "7.14": "7_14_IncidentOffenseStudentList",
        "7.15": "7_15_IncidentActionsPersistentlyDangerousOffenseExpulsions",
        "7.16": "7_16_IncidentRemovalsForStudentsWithDisabilitiesCount",
        "7.17": "7_17_UnilateralRemovalsForStudentsWithDisabilitiesCount",
        "7.18": "7_18_IncidentRemovalsForStudentsWithDisabilitiesStudentList",
        "8.1": "8_1_StudentProfileList_EOY3_",
        "14.1": "14_1_StudentAbsenteeismCount",
        "14.2": "14_2_StudentAbsencesStudentList",
    },
    "EOY4": {
        "16.1": "16_1_SPEDEducationPlanByPrimaryDisabilityCount_EOY4",
        "16.2": "16_2_SPEDPrimaryDisabilityByFederalSettingCount_EOY4",
        "16.3": "16_3_SPEDStudentProfileList_EOY4",
        "16.5": "16_5_StudentsWithDisabilitiesStudentServices_EOY4",
        "16.6": "16_6_StudentsWithDisabilitiesStudentServicesStudentList_EOY4",
        "16.9": "16_9_SPEDProgramExitCountbyPrimaryDisability_EOY4",
        "16.10": "16_10_SPEDProgramExitStudentList_EOY4",
        "16.11": "16_11_SPEDPrimaryDisabilityAnnualComparisionReport_EOY4",
        "17.3": "17_3_PostsecondarySurveyOutcomeforSWDsCount",
        "17.4": "17_4_PostsecondarySurveyOutcomeforSWDsStudentList",
    },
}
    